#!/usr/bin/perl

%families = (
#	'original'  => [ qw(original rt crc32 abcd64) ],
#	'abcd'	    => [ qw(abcd abcdef abcd_xor abcd64) ],
#	'rt'	    => [ qw(rt rt_ab rt_ab2 rt_new abcd64) ],
#	'crc32'	    => [ qw(crc32 rusty abcd64) ],
#	'guilmette' => [ qw(guilmette pearson abcd64) ],
#	'jenkins'   => [ qw(jenkins2 jenkins2w) ],
#	'abcdn'	    => [ qw(abcd-1 abcd-2 abcd-3 abcd-4) ],
	'jenkins'   => [ qw(jenkins2 jenkins3) ],
	'superfast' => [ qw(jenkins2 jenkins3 superfast) ],
	'murmur2'   => [ qw(jenkins2 jenkins3 murmur2) ],
);

@sizes = qw(8191 8209
	    16383 16411
	    32767 32771
	    65535 65537
	    131071 131101
	    262143 262147);
@even = qw(8192 16384 32768 65536 131072 262144);

@data = qw(50000 45000 40000 35000 30000 25000 20000 15000 10000);
push(@data, qw(9000 8000 7000 6000 5000 4000 3000 2000 1000));
push(@data, qw(900 800 700 600 500 400 300 200));
push(@data, reverse(1..100));

grep(push(@hashes, @{$families{$_}}), keys %families);
@hashes = grep(!$seen{$_}++, @hashes);

@commands = qw(cttest hashes stats html speed bit_test);

if (grep($ARGV[0] eq $_, @commands)) {
	eval $ARGV[0];
} else {
	&cttest;
	&hashes;
	&stats;
	&html;
	&speed;
}

sub cttest {
	system("gcc -o cttest -DPER_HASH_INNER -Os -march=native -lm cttest.c");
}

sub input {
	my($data) = shift;

	my $ip = 1;
	my $nw = 0;
	my $port = 1024;
	my $step = $data <= 100 ? 500 :
		$data <= 1000 ? 50 :
		$data <= 10000 ? 5 : 1;

        mkdir("stat.html") unless -d "stat.html";
	open(INPUT, ">stat.html/$data");
	while ($port < 51024) {
		my $addr = $ip;
		my $net = $nw;
		my($i, $j);
		for ($i = 0; $i < $data; $i++) {
			$addr += $i;
			if ($addr > 254) {
				$net++;
				$addr = $ip;
			}
			for ($j = 1; $j <= $step; $j++) {
				$port++;
				print INPUT "tcp\t6 255 ESTABLISHED src=192.168.$net.$addr dst=192.168.240.1 sport=$port dport=80 src=192.168.240.1 dst=192.168.$net.$addr sport=80 dport=$port\n";
			}
			last if $port >= 51024;
		}
	}
	close(INPUT);
}

sub hashes {
	my(@all, $data, $hash, $size);

	foreach $data (@data) {
		input($data);	
		foreach $hash (('random', @hashes)) {
			mkdir("stat.html/$hash");
			@all = @sizes;
			push(@all, @even);
				# if grep($hash eq $_, @{$families{best}}) || $hash eq 'random'; 
			print "$data $hash\n";
			foreach $size (@all) {
				next if -f "stat.html/$hash/$size.$data.txt";
				system("./cttest -h $hash -s $size -f stat.html/$data >stat.html/$hash/$size.$data.txt");
			}
		}
		unlink("stat.html/$data");
	}
}

sub stats {
	my(@all) = @sizes;

	push(@all, @even);
	@all = sort { $a <=> $b } @all;
	my($size, $data, $family, @random);
	my(%plot, @bucketlen, $title, $mean, $sdev, $adev, $max, $P, $chi2, $chi2rand, $walk);
	foreach $size (@all) {
		@random = ();
		foreach $data (@data) {
			open(IN, "stat.html/random/$size.$data.txt");
			while (<IN>) {
				if (/^(\d+)\s+(\d+)/) {
					# (count, length) = ($1, $2);
					$random[$2] = $1;
				}
			}
			close(IN);
		}
		foreach (0..$#random) {
			$random[$_] = 0 unless $random[$_];
			# $random[$_] /= $#data + 1;
		}
		foreach $family (keys %families) {
			next if $family eq 'random';
			# next if !($size % 2) && $family ne 'best';
			print "$family $size\n";
			%plot = ();
			$title = '';
			foreach $hash (@{$families{$family}}) {
				$title .= "'-' title \"$hash\" with linespoints, ";
				foreach $data (@data) {
					@bucketlen = ();
					open(IN, "stat.html/$hash/$size.$data.txt");
					while (<IN>) {
						if (/^maximum list length:\s+(\S+)/) {
							$plot{maxlen} .= "$1\n";
						} elsif (/longest bucket length:\s+(\S+)/) {
							$plot{probab} .= "$1\n";
						} elsif (/average list length:\s+(\S+)/) {
							$P = $1;
						} elsif (/^(\d+)\s+(\d+)/) {
							# (count, length) = ($1, $2);
							$bucketlen[$2] = $1;
						}
					}
					close(IN);
					$mean = $sdev = $adev = 0;
					foreach $len (0..$#bucketlen) {
						$bucketlen[$len] = 0 unless $bucketlen[$len];
						$mean += $bucketlen[$len];
					}
					$mean /= ($#bucketlen + 1);
					foreach $len (0..$#bucketlen) {
						$sdev += ($bucketlen[$len] - $mean)**2;
						$adev += abs($bucketlen[$len] - $mean);
					}
					$plot{sdev} .= $sdev/$#bucketlen . "\n";
					$plot{adev} .= $adev/($#bucketlen + 1) . "\n";
					$chi2 = 0;
					foreach $len (0..$#bucketlen) {
						$chi2 += $bucketlen[$len]*($len - $P)**2/$P;
					}
					$chi2 = ($chi2 - $size)/sqrt($size);
					$plot{chi2} .= $chi2 . "\n";
					$walk = 0;
					foreach $len (0..$#bucketlen) {
						$walk += $bucketlen[$len]*$len*($len +1)/2;
					}
					$plot{walk} .= $walk . "\n";
					$max = $#bucketlen > $#random ? 
					       $#bucketlen : $#random;
					$chi2rand = 0;
					foreach $len (0..$max) {
						$chi2rand += (($bucketlen[$len] - $random[$len])**2)/($bucketlen[$len] + $random[$len])
							unless $bucketlen[$len] == 0 && $random[$len] == 0;
					}
					$plot{chi2rand} .= $chi2rand . "\n";
				}
				foreach $what (qw(maxlen probab sdev adev chi2 chi2rand walk)) {
					$plot{$what} .= "e\n";
				}
			}
			foreach $data (@data) {
				$plot{chi2} .= "3\n";
			}
			$plot{chi2} .= "e\n";
			foreach $data (@data) {
				$plot{chi2} .= "-3\n";
			}
			$plot{chi2} .= "e\n";
			chop $title; chop $title;
			open(PLOT, "|gnuplot");
			print PLOT <<TXT;
cd "stat.html"
set terminal png small
set title "Hash size $size"
set ylabel "maximum list length"
set output "maxlen-$family-$size.png"
plot $title
$plot{maxlen}
set ylabel "standard deviation"
set output "sdev-$family-$size.png"
plot $title
$plot{sdev}
set ylabel "mean absolute deviation"
set output "adev-$family-$size.png"
plot $title
$plot{adev}
set ylabel "chi-square"
set output "chi2-$family-$size.png"
plot $title, '-' title "max" with linespoints, '-' title "min" with linespoints
$plot{chi2}
set ylabel "chi-square against random hash"
set output "chi2rand-$family-$size.png"
plot $title
$plot{chi2rand}
set ylabel "walking time"
set output "walk-$family-$size.png"
plot ${title}
$plot{walk}
set ylabel "probability"
set output "probab-$family-$size.png"
set logscale y 2
plot $title
$plot{probab}
TXT
			close(PLOT);
		}
	}
}

sub html {
	my(@all) = @sizes;

	push(@all, @even);
	@all = sort { $a <=> $b } @all;
	open(INDEX, ">stat.html/index.html");
	print INDEX <<TXT;
<head>
<title>Conntrack hash functions</title>
</head>
<body>

<h1>Conntrack hash function comparisons III</h1>
TXT
	open(PLOT, "|gnuplot");
	print PLOT <<TXT;
cd "stat.html"
set terminal png small
set ylabel "number of buckets"
set xlabel " bucket length"
set logscale y 10
set logscale x 2
TXT
	my($family, $foo, $html, $h, $size, $sample, $data, $plot);
	my(@title, @name, @hash);
	foreach $family (keys %families) {
		$foo = join(', ', @{$families{$family}});
		print INDEX <<TXT;
<h3>Hashes $foo</h3>
<ul>
<li><a href=\"maxlen.$family.html\">Maximum list lenght</a>
<li><a href=\"probab.$family.html\">Probability of one longest bucket length</a>
<li><a href=\"sdev.$family.html\">Standard devitation</a>
<li><a href=\"adev.$family.html\">Mean absolute deviation</a>
<li><a href=\"chi2.$family.html\">Chi-square</a>
<li><a href=\"chi2rand.$family.html\">Chi-square against random hash</a>
<li><a href=\"walk.$family.html\">Walking time</a>
</ul>
TXT
		$html = <<TXT;
<head>
<title>TITLE</title>
</head>
<body>

<h1>TITLE</h1>
TXT
		foreach $size (@all) {
			next if !($size % 2) && $family ne 'best';
			$html .= <<TXT;
<h3><a name="$size">Hash size $size</a></h3>
<img src="NAME-$family-$size.png"><br>
<a href="sample-$family-$size.html">Sample graphs</a>
TXT
			$sample = <<TXT;
<head>
<title>Sample graphs</title>
</head>
<body>

<h1>Sample graphs: $foo, hash size $size</h1>
TXT
			# Samples
			foreach $data (qw(50000 10000 1000)) {
				$plot = '';
				@hash = @{$families{$family}};
				push(@hash, 'random');
				foreach $hash (@hash) {
					$plot .= "'$hash/$size.$data.txt' using 2:1 title \"$hash\" with linespoints, ";
				}
				chop($plot); chop($plot);
				$sample .= <<TXT;
<h3><a name="$data">Sample input: hash size $size, $data clients</a></h3>
<img src="sample-$family-$size-$data.png">
TXT
				print PLOT <<TXT;
set title "Sample input: $data"
set output "sample-$family-$size-$data.png"
plot $plot
TXT
			}
			$sample .= "</body>\n";
			open(HTML, ">stat.html/sample-$family-$size.html");
			print HTML $sample;
			close(HTML);
		}
		$html .= "</body>\n";
		@title = ('Maximum list length',
			  'Probability of one longest bucket length',
			  'Standard deviation',
			  'Mean absolute deviation',
			  'Chi-square',
			  'Chi-square against random hash',
			  'Walking time');
		@name = qw(maxlen probab sdev adev chi2 chi2rand walk);
		foreach (0..$#name) {
			($h = $html) =~ s/TITLE/$title[$_]: $foo/g;
			$h =~ s/NAME/$name[$_]/g;
			open(HTML, ">stat.html/$name[$_].$family.html");
			print HTML $h;
			close(HTML);
		}
	}
	print INDEX <<TXT;
<h3>Speed comparisons</h3>
<img src="timeplot-all-hashes.png">
TXT
	if ($#hashes > 12) {
		print INDEX <<TXT
<img src="timeplot-hashes.png">
</body>
TXT
	}
	close(INDEX);
	close(PLOT);
}

sub speed {
	my($hash, $n, $m, $l, $plot, $data, $plot2, $data2);
	foreach $hash (@hashes) {
		next if $hash =~ /-\d+$/;
		open(IN, "stat.html/$hash/16383.1.txt");
		while (<IN>) {
			if (/\s+(\d+) cyc,/) {
				$hashes{$hash} = $1;
				close(IN);
			}
		}
	}

	@hashes = sort { $hashes{$a} <=> $hashes{$b} } keys %hashes;
	$n = 0;
	$m = 2;
	$l = 0;
	foreach (@hashes) {
		$plot .= "'-' title \"$_\" with linespoints, ";
		$data .= "$n $hashes{$_}\n$m $hashes{$_}\ne\n";
		if ($l <= 12) {
			$plot2 .= "'-' title \"$_\" with linespoints, ";
			$data2 .= "$n $hashes{$_}\n$m $hashes{$_}\ne\n";
		}
		$n += 2;
		$m += 2;
		$l++;
	}
	chop $plot; chop $plot;
	chop $plot2; chop $plot2;
	open(PLOT, "|gnuplot");
	print PLOT <<TXT;
cd "stat.html"
set terminal png small
set output "timeplot-all-hashes.png"
set ylabel "cyc"
set key outside
plot $plot
$data
set output "timeplot-hashes.png"
plot $plot2
$data2
TXT
	close(PLOT);
}

sub bit_test {
	foreach $hash (@hashes) {
		system("./cttest -h $hash -B");
	}
}
